// ShmMgr.cpp: implementation of the CShmMgr class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "ServerSecurityAttribute.h"

//
// PROJECT INCLUDE
//
#include "LogDef.h"
#include <cmcl.h>
#include <exception>
#include "ShmMgr.h"
#include "ShmLoggerMgr.h"
 
#include "FragmentStringTable.h"

using namespace std;


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
CShmLogMgr::~CShmLogMgr()
{
	delete m_pLogAppender;
	//
	// All other members can be auto deleted
	//
}
  

CShmLogMgr::CShmLogMgr(	bool bIsServer, 
						PVOID MemoryAddress, 
						ULONG MemorySize,
						std::string& KeyPrefix,
						CAppender*pAppender):
m_ShmBuffer((LOGBUFFER*)MemoryAddress),
m_MaxSize(MemorySize),
m_strKeyPrefix(KeyPrefix),
m_LogEntrySemPtr(new CMclSemaphore(0,
								(MemorySize-HeaderSize)/LOGENTRY_STRING_SIZE,
								(KeyPrefix+SEMAPHOE_POSTFIX_PRODUCT).c_str(),
								CServerSecurityAttribute::RealNullOne().GetNewsSecurityAttribute())),
m_LogSpaceSemPtr(new CMclSemaphore((MemorySize-HeaderSize)/LOGENTRY_STRING_SIZE,
								(MemorySize-HeaderSize)/LOGENTRY_STRING_SIZE,
								(KeyPrefix+SEMAPHOE_POSTFIX_AVAIL).c_str(),
								CServerSecurityAttribute::RealNullOne().GetNewsSecurityAttribute()))

 
{
	m_bIsValid=	CMclIsValidHandle((*m_LogEntrySemPtr).GetHandle())&&
				CMclIsValidHandle((*m_LogSpaceSemPtr).GetHandle())&&
				CMclIsValidHandle(m_StopNotify.GetHandle())&&
				CMclIsValidHandle(m_StopAck.GetHandle()); 

	if (m_bIsValid)
	{
		m_bIsServer=true;
		m_ShmBuffer->count=0;
		m_ShmBuffer->MaxCount=(MemorySize-HeaderSize)/LOGENTRY_STRING_SIZE;
		m_ShmBuffer->read=-1;
		m_ShmBuffer->write=-1;
		m_ShmBuffer->ulSig=0xBEEF;
	}
	if (!m_bIsValid)
	{
		OutputDebugStringA("Unable to create the kernal object");
	}

	m_bStop=true;

	m_pLogAppender =pAppender;

}

CShmLogMgr::CShmLogMgr(	PVOID MemoryAddress, 
						ULONG MemorySize,
						std::string& KeyPrefix):
m_ShmBuffer((LOGBUFFER*)MemoryAddress),
m_MaxSize(MemorySize),
m_strKeyPrefix(KeyPrefix),
m_LogEntrySemPtr(new CMclSemaphore((KeyPrefix+SEMAPHOE_POSTFIX_PRODUCT).c_str())),
m_LogSpaceSemPtr(new CMclSemaphore((KeyPrefix+SEMAPHOE_POSTFIX_AVAIL).c_str()))
{
	m_bIsValid=	CMclIsValidHandle((*m_LogEntrySemPtr).GetHandle())&&
				CMclIsValidHandle((*m_LogSpaceSemPtr).GetHandle())&&
				CMclIsValidHandle(m_StopNotify.GetHandle())&&
				CMclIsValidHandle(m_StopAck.GetHandle()); 

	if (m_bIsValid)
	{
		if(InterlockedCompareExchange((LONG*)(&m_ShmBuffer->ulSig),
						0XBEEF, 0XBEEF) == 0XBEEF) 
		{
			m_bIsValid=true;
		}
		else
		{
			m_bIsValid=false;
		}
	}

	if (!m_bIsValid)
	{
		OutputDebugStringA("Unable to create the kernal object");
	}
		
	m_bIsServer=false;
	m_bStop=true;

}


CShmLogMgr* CShmLogMgr::CreateServer(PVOID addr, 
									ULONG size, 
									std::string& strPrefix,
									CAppender* pAppender)
{
	if (addr&&size)
	{
		CShmLogMgr* p= new CShmLogMgr(true,addr,size,strPrefix,pAppender);
		if (p)
		{
			if (*p)
			{
				return p;
			}
			else
			{
				p->Delete();
			}

		}
	}

	return NULL;
}

CShmLogMgr* CShmLogMgr::CreateClient(PVOID addr, ULONG size, std::string& strPrefix)
{
	CShmLogMgr* p= new CShmLogMgr(addr,size,strPrefix);

	if (p)
	{
		if (*p)
		{
			return p;
		}
		else
		{
			p->Delete();
		}
	}
	return NULL;
}

bool CShmLogMgr::WriteString (std::vector<char>& data)
{
	if (! *this)
	{
		return false;
	}

	int totalSize=data.size();
	if (totalSize<EntryPayloadSize)
	{
		return WriteToBuffer(data, false, true);
	}


	int currentStartPoint=0;
	while (totalSize>EntryPayloadSize)
	{
		std::vector<char> writeData;
		writeData.resize(EntryPayloadSize);
		for(int i=0; i<EntryPayloadSize;i++)
		{
			writeData[i]=data[currentStartPoint+i];
		}
		currentStartPoint+=i;
		totalSize-=EntryPayloadSize;
		if (!WriteToBuffer(writeData,true,false))
		{
			return false;
		}
 	}

	std::vector<char> writeData;
	writeData.resize(totalSize);
		
	for(int i=0; i<totalSize;i++)
	{
		writeData[i]=data[currentStartPoint+i];
	}
	return WriteToBuffer(writeData,true,true);

}


bool CShmLogMgr::WriteToBuffer(std::vector<char>& data,
							   bool bFragmented,
							   bool endFragment)
{
	if (! *this)
	{
		return false;
	}

	if (data.size()>EntryPayloadSize)
	{
		return false;
	}

	DWORD waitRet=m_LogSpaceSemPtr->Wait(1000);
	if (CMclWaitSucceededIndex(waitRet)!=WAIT_OBJECT_0)
	{
		//
		// If no space available in one second. The CLM service is considered deadth.
		// we won't log anything afterwards.
		//
		m_bIsValid=false;
		return false;
	}

	// Get a big enough slot in the ring buffer
	if(InterlockedIncrement(&m_ShmBuffer->count) > m_ShmBuffer->MaxCount)
	{
		//
		// The memory is full
		//
		InterlockedDecrement(&m_ShmBuffer->count);
		return false;
	}

	DWORD write_location = InterlockedIncrement(&m_ShmBuffer->write);
	write_location%=m_ShmBuffer->MaxCount;


	PLOG_ENTRY plogEntry=(PLOG_ENTRY) (m_ShmBuffer->buffer+LOGENTRY_STRING_SIZE*write_location);

	plogEntry->size=EntryHeaderSize+data.size();
	plogEntry->flag=0;
	if (bFragmented)
	{
		plogEntry->flag|=0x80;
		if (!endFragment)
		{
			plogEntry->flag|=0x40;
		}
		else
		{
			plogEntry->flag=0x80;

		}
	}
	plogEntry->Handle=GetCurrentProcessId();
	std::copy(data.begin(), data.end(), plogEntry->buffer);

	
	LONG preCount=0;
	m_LogEntrySemPtr->Release(1,&preCount);

	return true;
}


bool CShmLogMgr::ReadOneEntry(std::vector<char>& entry)
{
	if (! *this)
	{
		return false;
	}

	if(InterlockedCompareExchange(&m_ShmBuffer->count, 0, 0) == 0) return false;


    m_ShmBuffer->read=(m_ShmBuffer->read+1)%m_ShmBuffer->MaxCount;

	PLOG_ENTRY plogEntry=
		(PLOG_ENTRY) (m_ShmBuffer->buffer+LOGENTRY_STRING_SIZE*m_ShmBuffer->read);

	entry.resize(plogEntry->size);
	
	if (plogEntry->size<=LOGENTRY_STRING_SIZE)
	{
		std::copy((char*)plogEntry, (char*)plogEntry+plogEntry->size, entry.begin());
	}
 
	InterlockedDecrement(&m_ShmBuffer->count);

	LONG prev;
	m_LogSpaceSemPtr->Release(1,&prev);	
	
  	return true;

}

void CShmLogMgr::SetLogAppender(CAppender* p)
{
	if (m_pLogAppender)
	{
		delete m_pLogAppender;
	}
	m_pLogAppender=p;
}


void CShmLogMgr::ProcessLogs()
{
	CLMAssert(m_bIsValid, "The logger manager is invalid");
	CLMAssert(m_bIsServer, "The non-server logger was trying to process log entry");

	m_bStop =false;

	std::vector<char> logEntry;

	while (!m_bStop)
	{
		DWORD status=m_StopNotify.WaitForTwo
					(*m_LogEntrySemPtr,FALSE,INFINITE );

		if (CMclWaitSucceededIndex(status)==WAIT_OBJECT_0+1)
		{
			//
			// We got one element
			//
			if(ReadOneEntry(logEntry))
			{
				ProcessEntry((PLOG_ENTRY)logEntry.begin());
			}
		}
		else
		{
			m_bStop=true;
		}
	}

	m_StopAck.Set();
	
}

void CShmLogMgr::StopLogger()
{
	if (!m_bStop)
	{
		m_bStop=true;
		m_StopNotify.Set();
		m_StopAck.Wait(5000);
	}
}


void CShmLogMgr::ProcessEntry(PLOG_ENTRY pEntry)
{
	if (pEntry->flag&0x80)
	{
		if(!CFragStringTable::Instance_DebugLog().
					AddString(pEntry->Handle,
							  pEntry->buffer,
							  pEntry->size-EntryHeaderSize))
		{
			return;
		}

		if (!(pEntry->flag&0x40))
		{
			//
			// No more data call the special processor
			//
			vector<char> retBuffer;
 			if(CFragStringTable::Instance_DebugLog().
				GetString(pEntry->Handle,retBuffer))
			{
				retBuffer.push_back('\0');
				if (m_pLogAppender)
				{
					m_pLogAppender->AppendLogInfo(retBuffer.begin());
				}
			}
		}
	}
	else
	{
		vector<char>retBuffer;
		int payloadSize=pEntry->size-EntryHeaderSize;
		retBuffer.resize(payloadSize);
		std::copy(pEntry->buffer,pEntry->buffer+payloadSize,retBuffer.begin());
		retBuffer.push_back('\0');
		if (m_pLogAppender)
		{
			m_pLogAppender->AppendLogInfo(retBuffer.begin());
		}
	}

}



